#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include "mmc.h"
#include "spi.h"
#include "delay.h"
#include "fat16.h"
#include "types.h"

//#define RESCUE

#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega88__)

//mega168
#define UDR UDR0
#define UCSRA UCSR0A
#define UCSRB UCSR0B
#define UCSRC UCSR0C
#define UBRRL UBRR0L
#define UBRRH UBRR0H
#define UBRR UBRR0
#define UDRE UDRE0
#define RXEN RXEN0
#define TXEN TXEN0
#define RXCIE RXCIE0
#define UDRIE UDRIE0
#define OCR2 OCR2A
#define TCCR0 TCCR0B
#define TCCR2 TCCR2B
#endif

#ifndef F_CPU
	#define F_CPU 8000000
#endif

// Buttons
#define KEY_PORT PIND
#define KEY_MASK 0x08

// Pins
#define PORT_LED PORTB
#define PIN_LED 2 //PORTB

// States
#define STOP 0
#define REC 2

// MMC
#define START_SECTOR 528	// This is the sector number where the search for midi files begins (no FAT32 readings!)
#define TX_BUF_SIZE 128
#define RX_BUF_SIZE 128

// Variables
u08 sharedmem[512];
u32 time;
u08 tx_buf[TX_BUF_SIZE], rx_buf[RX_BUF_SIZE], rx_wrp, rx_rdp, tx_wrp, tx_rdp; // TX buffer
u08 state=STOP, sample=0, channels=1, uartrx;	// Flags
u08 file_cnt=0;	// File
u08 gl_len=0, gl_pos=0, gl_error, adcflag;



extern struct fat_filedata_st fat_filedata;
extern u32 file_len;
extern u08 fat_cardfull;


// Local Functions
static void hw_init(void);


//---------------------------------------------------

static void error(void) {
	do {
		PORT_LED &= ~PIN_LED;
		delay_ms(100);
		PORT_LED |= PIN_LED;
		delay_ms(100);
	} while (1);
}


// Init MCU
static void hw_init(void) {
//	u08 k;

	// Oscillator
	// Ports
	DDRC = 0x00;	// adc
	PORTC = 0x00;

	DDRB = 0x2F;
	PORTB = 0xFF-2;

	DDRD = 0x12;	// Input-Port, auer TxD
	PORTD = 0xEF;	// Pull-Ups aktivieren

	//Initialisierung der MMC/SD-Karte
	MMC_hw_init();
	spi_init();
	delay_ms(100);
	sei();

	if (!MMC_init()) {
		error();
	}

#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega88__)
	TIMSK1 = 2;
#else
	TIMSK = 1<<OCIE1A;
#endif
}



static void setbaud(u16 b) {
	UCSRB = 0;
	if (b >= 1200) {
		UCSRA = 0x02;
#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega88__)
		UBRR = F_CPU/8/b-1;
#else
		UBRRL = (F_CPU/8/b-1);
		UBRRH = (F_CPU/8/b-1)>>8;
#endif
	} else {
		UCSRA = 0;
#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega88__)
		UBRR = F_CPU/16/b-1;
#else
		UBRRL = (F_CPU/16/b-1);
		UBRRH = (F_CPU/16/b-1)>>8;
#endif
	}
	UCSRB = (1<<RXEN)|(1<<TXEN);
}


// Send UART byte
static void sendchar(u08 byte) {
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = byte;
}


static void send_ee() {
//	u08 i=0, l=eeprom_read_byte(0);
//	while (i<l)
//		sendchar(eeprom_read_byte(1+i++));
	gl_pos = 0;
	if (gl_len)
		UCSRB |= 1<<UDRIE;
}


// Uhr starten
/*static void start_time(u32 interval) {
	TCNT1 = 0;
	time = 0;
	if (interval < 65536UL) {
		OCR1A = interval;
		TCCR1B = 8+1;
	}
	else if (interval < (u32)65536UL*8UL) {
		OCR1A = interval/8;
		TCCR1B = 8+2;
	}
	else if (interval < (u32)65536UL*64UL) {
		OCR1A = interval/64;
		TCCR1B = 8+3;
	}
	else if (interval < (u32)65536UL*256UL) {
		OCR1A = interval/256;
		TCCR1B = 8+4;
	}
	else {
		OCR1A = interval/1024;
		TCCR1B = 8+5;
	}
}
*/

// Uhr stoppen
inline static void stop_time(void) {
	TCCR1B = 0;
}


static inline void write_nibble(u08 n) {
	if (n<10)
		mmc_write_byte(n+'0');
	else
		mmc_write_byte(n-10+'A');
}


static inline void write_hex_u08(u08 n) {
	write_nibble(n>>4);
	write_nibble(n&0x0f);
}


static u08 check_nibble(u08 n) {
	if (n<='9' && n>='0')
		return 1;
	else if (n<='F' && n>='A')
		return 1;
	else if (n<='f' && n>='a')
		return 1;
	return 0;
}


static u08 nibble_to_bin(u08 n) {
	if (n<='9')
		return (n-'0');
	else if (n<='F')
		return (n-'A'+10);
	else if (n<='f')
		return (n-'a'+10);
	return 0;
}


static inline u08 hex2bin_u08(u08 n1, u08 n2) {
	return (nibble_to_bin(n1)<<4) | nibble_to_bin(n2);
}


// Dezimale Zahlen 0..65536
static void write_number_u32(u32 d) {
	u08 zehner=0, hunderter=0, tausender=0, zehntausender=0, hunderttausend=0, millionen=0, zehnmillionen=0, hundertmillionen=0, milliarden=0;
	while (d >= 1000000000) {
		d -= 1000000000;
		milliarden++;
	}
	while (d >= 100000000) {
		d -= 100000000;
		hundertmillionen++;
	}
	while (d >= 10000000) {
		d -= 10000000;
		zehnmillionen++;
	}
	while (d >= 1000000) {
		d -= 1000000;
		millionen++;
	}
	while (d >= 100000) {
		d -= 100000;
		hunderttausend++;
	}
	while (d >= 10000) {
		d -= 10000;
		zehntausender++;
	}
	while (d >= 1000) {
		d -= 1000;
		tausender++;
	}
	while (d >= 100) {
		d -= 100;
		hunderter++;
	}
	while (d >= 10) {
		d -= 10;
		zehner++;
	}
	u08 printzero=0;
	if (milliarden) {
		mmc_write_byte(milliarden+'0');
		printzero = 1;
	}
	if (hundertmillionen || printzero) {
		mmc_write_byte(hundertmillionen+'0');
		printzero = 1;
	}
	if (zehnmillionen || printzero) {
		mmc_write_byte(zehnmillionen+'0');
		printzero = 1;
	}
	if (millionen || printzero) {
		mmc_write_byte(millionen+'0');
		printzero = 1;
	}
	if (hunderttausend || printzero) {
		mmc_write_byte(hunderttausend+'0');
		printzero = 1;
	}
	if (zehntausender || printzero) {
		mmc_write_byte(zehntausender+'0');
		printzero = 1;
	}
	if (tausender || printzero) {
		mmc_write_byte(tausender+'0');
		printzero = 1;
	}
	if (hunderter || printzero) {
		mmc_write_byte(hunderter+'0');
		printzero = 1;
	}
	if (zehner || printzero) {
		mmc_write_byte(zehner+'0');
		printzero = 1;
	}
	mmc_write_byte(d+'0');
}


static s08 str_ncomp(u08 *buf, u08 *buf2, u08 len) {
	u08 cnt=0;

	for (cnt=0; cnt<len; cnt++) {
		if (buf[cnt] < buf2[cnt])
			return 1;
		else if (buf[cnt] > buf2[cnt])
			return -1;
	}
	return 0;
}


// Read string from MMC and verify content
static u08 eat_string(u08 *buf, u08 len) {
	u08 b, cnt=0;

	for (cnt=0; cnt<len; cnt++) {
		do {
			b = mmc_fetch_byte();
		} while (b == 13 || b == 10);
		if (b != (u08)buf[cnt])
			return 0;
	}
	return 1;
}


static u16 eat_number_u16(void) {
	u08 b;
	u16 erg=0;

	while (1) {
		b = mmc_fetch_byte();
		if (b>='0' && b<='9') {
			erg *= 10;
			erg += b-'0';
		}
		else
			return erg;
	}
}


//----------------------------------------------------------
// Main
#ifdef RESCUE
	extern u08 fat_directaccess;
#endif

void main(void) {
	u08 i, cyclic=0, binary=0;
	u16 overflow=781/*2*/, k, baud=0;
	u08 prescaler=8+5;

	hw_init();
	PORT_LED |= PIN_LED;
	delay_ms(100);
	PORT_LED &= ~PIN_LED;
#ifdef RESCUE
	setbaud(38400);
	tx_rdp = 0;
	tx_wrp = 0;
	uartrx = 0;
	fat_directaccess=1;
	mmc_write_start(0);
	u32 p=0;
	while (1) {
		if (uartrx) {
			uartrx=0;
			while (tx_rdp != tx_wrp) {
				mmc_write_byte(tx_buf[tx_rdp++]);
				p++;
			}
		}
		if (p >= 320000) {
			PORT_LED |= PIN_LED;
			mmc_complete_write();
			while (1);
		}
	}
#endif
	if (!fat_init()) {
		error();
	}
	else {
		fat_count_files();
		if (file_cnt) {
			for (k=0; k<file_cnt; k++) {
				fat_read_filedata(k);
				if (str_ncomp(fat_filedata.name, "SETTINGS.TXT", 12) == 0 || str_ncomp(fat_filedata.name, "settings.txt", 12) == 0) {
					//mmc_read_sector(fat_filedata.startsect, sharedmem);
					fat_openfile(FILEMODE_READ, k);
					if (eat_string("ch=", 3)) {
						channels = eat_number_u16();
						if (channels > 6)
							channels = 6;
					}
					if (eat_string("overflow=", 9)) {
						overflow = eat_number_u16();
					}
					if (eat_string("prescaler=", 10)) {
						prescaler = eat_number_u16();
						//if (prescaler == 0)
						//	prescaler = 8+1;
					}
					if (eat_string("baud=", 5)) {
						baud = eat_number_u16();
						if (baud > 57600)
							baud = 57600;
						else if (baud < 300)
							baud = 300;
						setbaud(baud);
					}
					if (baud && eat_string("binary=", 7)) {
						binary = eat_number_u16();
					}
					if (baud && eat_string("send=", 5)) {
						while (1) {
							u08 n1, n2;
							n1 = mmc_fetch_byte();
							n2 = mmc_fetch_byte();
							if (check_nibble(n1) && check_nibble(n2))
								sendchar(hex2bin_u08(n1, n2));
							else
								break;
						}
					}
					if (baud && eat_string("cyclic=", 7)) {
						cyclic = 1;
						i = 0;
						while (1) {
							u08 n1, n2;
							n1 = mmc_fetch_byte();
							n2 = mmc_fetch_byte();
							if (check_nibble(n1) && check_nibble(n2))
								eeprom_write_byte(i++, hex2bin_u08(n1, n2));
							else {
								gl_len = i;
								break;
							}
						}
					}
					fat_closefile();
				}
			}
		}
	}
	PORT_LED |= PIN_LED;

	while (1) {
		switch (state) {
			case STOP:
				if ((u08)(~KEY_PORT) & KEY_MASK) {
					delay_ms(100);
					state = REC;
					PORT_LED &= ~PIN_LED;
					while ((u08)(~KEY_PORT) & KEY_MASK);
				}
				break;

			case REC:
				// Rekorder-Task
				fat_openfile(FILEMODE_WRITE, file_cnt);
				if (fat_cardfull) {
					state = STOP;
					PORT_LED |= PIN_LED;
				}
				mmc_write_start(fat_filedata.startsect);
				file_len = 0;
			
				tx_rdp = 0;
				rx_rdp = 0;
				tx_wrp = 0;
				rx_wrp = 0;
				sample = 0;
				gl_error = 0;
				adcflag = 0;
				uartrx = 0;
				//start_time((u32)(interval*F_CPU));
				time = 0;
				OCR1A = overflow;
				TCNT1 = overflow-1;
				TCCR1B = prescaler;

				u08 linebrkneeded=0;
				if (baud)
					UCSRB |= 1<<RXCIE;
				while (1) {
					if (uartrx) {
						uartrx = 0;
						while (rx_rdp != rx_wrp) {
							if (binary)
								mmc_write_byte(rx_buf[rx_rdp++]);
							else
								write_hex_u08(rx_buf[rx_rdp++]);
							rx_rdp &= RX_BUF_SIZE-1;
						}
						linebrkneeded = 1;
					}
					else if (channels && sample == channels) {
						adcflag = 0;
						sample = 0;
						if (linebrkneeded) {
							linebrkneeded = 0;
							mmc_write_byte(0x0d);
							mmc_write_byte(0x0a);
						}
						while (tx_rdp != tx_wrp) {
							mmc_write_byte(':'); write_number_u32(time);
							for (i=0; i<channels; i++) {
								mmc_write_byte(9);
								write_number_u32(tx_buf[tx_rdp]+256*tx_buf[tx_rdp+1]);
								tx_rdp+=2;
								tx_rdp &= TX_BUF_SIZE-1;
							}
							mmc_write_byte(0x0d);
							mmc_write_byte(0x0a);
						}
						if (cyclic)
							send_ee();
						PORT_LED &= ~PIN_LED;
					}
					else if (!channels && sample == 1) {
						sample = 0;
						if (linebrkneeded) {
							linebrkneeded = 0;
							mmc_write_byte(0x0d);
							mmc_write_byte(0x0a);
						}
						mmc_write_byte(':'); write_number_u32(time);
						mmc_write_byte(0x0d);
						mmc_write_byte(0x0a);
						if (cyclic)
							send_ee();
						PORT_LED &= ~PIN_LED;
					}
					if (gl_error || ((u08)(~KEY_PORT) & KEY_MASK) || fat_cardfull) {
						stop_time();
						if (baud)
							UCSRB &= ~(1<<RXCIE);
						fat_filedata.len = file_len;
						mmc_complete_write();
						fat_closefile();
						file_cnt++;
						PORT_LED |= PIN_LED;
						delay_ms(100);
						state = STOP;
						while ((u08)(~KEY_PORT) & KEY_MASK);
						break;
					}
				}
				
				if (gl_error)
					error();

				break;
		}
	}
}


//----------------------------------------------------------
// IRQ

ISR(SIG_OUTPUT_COMPARE1A) {
	PORT_LED |= PIN_LED;
	time++;
	if (channels) {
		sample = 0;
		// ADC start
		ADMUX = 0x40+0;
		ADCSRA = 0xc8+6;
	}
	else
		sample = 1;
}

//volatile u16 adcval;

ISR(SIG_ADC) {
//	adcval = ADC;
	if (adcflag && tx_wrp==tx_rdp) {
		gl_error = 1;
#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega88__)
		TIMSK1 = 0;
#else
		TIMSK = 0;
#endif
	}
	else {
		tx_buf[tx_wrp++] = ADCL;
		tx_buf[tx_wrp++] = ADCH;
		tx_wrp &= TX_BUF_SIZE-1;
		adcflag = 1;
		sample++;
		if (sample < channels) {
			ADMUX++;
			ADCSRA = 0xc8+6;
		}
	}
}

#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega88__)
  ISR(SIG_USART_RECV) {
#else
  ISR(SIG_UART_RECV) {
#endif
#ifndef RESCUE
	if (state == STOP)
		return;
	if (uartrx && rx_wrp==rx_rdp) {
		gl_error = 1;
		UCSRB &= ~(1<<RXCIE);
	}
	else {
		rx_buf[rx_wrp++] = UDR;
		rx_wrp &= RX_BUF_SIZE-1;
#else
	{
		tx_buf[tx_wrp++] = UDR;
#endif
		uartrx = 1;
	}
}


#if defined(__AVR_ATmega168__) || defined(__AVR_ATmega88__)
  ISR(SIG_USART_DATA) {
#else
  ISR(SIG_UART_DATA) {
#endif
	UDR = eeprom_read_byte(gl_pos++);
	if (gl_pos == gl_len)
		UCSRB &= ~(1<<UDRIE);
}
